from pathlib import Path
import importlib.resources as pkg_resources
from configparser import ConfigParser

class GraphRef:

    def __init__(self, terms, rels, rels_id):
        self._terms = terms
        self._rels = rels
        self._rels_id = rels_id
        self._config = None

        self._read_config()

    def _read_config(self):
        config = ConfigParser()

        current_file_folder = Path(__file__)
        # retrieve configuration file from the grandparent directory
        config_path = current_file_folder.parent.parent /  "config.ini"

        if config_path.exists():
            config.read(config_path)
        else:
            with pkg_resources.open_text('cemento', 'config.ini') as config_file:
                config.read_file(config_file)

        self._config = config

    def get_term(self, id):
        return self._terms[id]
    
    def get_rel(self, id):
        return self._rels[id]
    
    def get_rel_from_edge(self, edge):
        return self._rels[self._rels_id[edge]]
    
    def get_rel_id(self, edge):
        try:
            return self._rels_id[edge]
        except KeyError:
            return None


import networkx as nx
import pandas as pd
from cemento.draw_io.diagram_ref import DiagramRef
from cemento.graphref import GraphRef

class Graph:

    def __init__(
        self,
        file_path=None,
        rels_df=None,
        graph=None,
        ref=None,
        do_gen_ids=False,
        infer_rank=False,
    ):
        self._file_path = file_path
        self._rels_df = rels_df
        self._graph = graph
        self._ref = ref
        self._edges = None
        self._nodes = None
        self._predicate_edges = None
        self._rank_edges = None

        if (file_path is not None) + (graph is not None) + (rels_df is not None) != 1:
            raise TypeError(
                "Too many or too few arguments were provided. Either use file_path, graph or rels_df."
            )

        if file_path or rels_df is not None:
            if file_path:
                terms, rels, rels_id, edges = self._read_edges(
                    do_gen_ids, infer_rank, file_path=self._file_path
                )
            elif rels_df is not None:
                terms, rels, rels_id, edges = self._read_edges(
                    do_gen_ids, infer_rank, rels_df=self._rels_df
                )

            self._ref = GraphRef(terms, rels, rels_id)

            self._predicate_edges = [edge for edge in edges if edge[2] == "predicate"]
            self._rank_edges = [edge for edge in edges if edge[2] == "rank"]
            unclassifed = [edge for edge in edges if edge[2] is None]

            self._edges = self.get_predicate_edges() + self.get_rank_edges() + unclassifed
            self._graph = self._create_graph(self._edges)

        if graph:
            self._edges = list(graph.edges())
            if ref:
                self._ref = ref

        self._nodes = self._graph.nodes()

    def _read_edges(self, do_gen_ids, infer_rank, file_path=None, rels_df=None):

        if file_path:
            df = pd.read_csv(file_path)

        if rels_df is not None:
            df = rels_df

        if not do_gen_ids and ("parent_id" not in df or "child_id" not in df or "rel_id" not in df):
            raise AttributeError(
                "Not enough information in file. Consider using the gen_ids option or creating the parent_id, child_id or rel_id column in your file"
            )

        if do_gen_ids:
            unique_terms = list(df["parent"].unique()) + list(df["child"].unique())
            unique_term_set = set(unique_terms)
            term_id = {
                unique_term: idx for idx, unique_term in enumerate(unique_term_set)
            }

            df["parent_id"] = df["parent"].map(term_id)
            df["child_id"] = df["child"].map(term_id)
            df["comb_id_temp"] = (
                df["parent_id"].astype("string") + "_" + df["child_id"].astype("string")
            )
            df["rel_id"] = pd.factorize(df["comb_id_temp"])[0]

            df.drop("comb_id_temp", axis=1, inplace=True)

        if infer_rank:
            temp_diagram_ref = DiagramRef()
            df["is_rank"] = df["rel"].map(temp_diagram_ref.is_rank)

        terms = {term_id: term for term_id, term in zip(df["parent_id"], df["parent"])}
        terms.update(
            {term_id: term for term_id, term in zip(df["child_id"], df["child"])}
        )
        rels = {edge_id: edge for edge_id, edge in zip(df["rel_id"], df["rel"])}
        rels_id = {
            (parent_id, child_id): rel_id
            for parent_id, child_id, rel_id in zip(
                df["parent_id"], df["child_id"], df["rel_id"]
            )
        }

        if "is_rank" in df:
            edges = [(parent, child, "rank" if is_rank else "predicate") for parent, child, is_rank in zip(df["parent_id"], df["child_id"], df["is_rank"])]
        else:
            edges = [(parent, child, None) for parent, child in zip(df["parent_id"], df["child_id"])]

        return terms, rels, rels_id, edges

    def _create_graph(self, edges):
        graph = nx.DiGraph()

        for u, v, edge_type in edges:
            graph.add_edge(u, v, type=edge_type)

        return graph

    def set_ref(self, ref):
        self._ref = ref

    def get_rels_df(self):
        if self._rels_df is None:
            raise ValueError("the relationship dataframe is not set. Please consider setting or reinitializing the object.")
        return self._rels_df

    def set_attr(self, node, attr, value):
        self._graph.nodes[node][attr] = value

    def get_graph(self):
        return self._graph

    def get_nodes(self):
        if self._nodes:
            return self._nodes
        return []

    def add_node(self, new_node_id):
        self._graph.add_node(new_node_id)

    def get_ref(self):
        return self._ref

    def get_edges(self):
        if self._edges:
            return self._edges
        return []

    def get_predicate_edges(self):
        if self._predicate_edges:
            return self._predicate_edges
        return []

    def get_rank_edges(self):
        if self._rank_edges:
            return self._rank_edges
        return []

    def get_attr(self, node, attr):
        return self._graph.nodes[node][attr]


from os import path
from defusedxml import ElementTree as ET

class WriteErrorDiagram:

    def __init__(self, read_diagram, print_id_width=25, separator_width=75):
        self._read_diagram = read_diagram
        self._print_id_width = print_id_width
        self._separator_width = separator_width

    def add_error_highlighting(self):
        tree = ET.parse(self.get_read_diagram().get_file_path())
        root = tree.getroot()

        errors = self.get_read_diagram().get_errors()

        for elem_id, _ in errors.items():
            for element in root.findall(f".//*[@id='{elem_id}']"):
                # read and strip element styles and save as pairs
                styles = list(map(lambda x: x.strip().split('='), element.get('style').strip().split(';')))
                styles = [style for style in styles if style[0]]

                # convert style pairs into dictionary and add/modify strokeColor entry
                style_dict = dict()
                for style in styles:
                    style_dict[style[0]] = '' if len(style) < 2 else  style[1]
                style_dict['strokeColor'] = "#ff0000"
            
                # save the new styles and set to element
                new_style = [f"{key}={value}" if value else f"{key}" for key, value in style_dict.items()]
                new_style = ';'.join(new_style)

                # set new style to the element attribute
                element.set('style', new_style)

        # save the file as a checked version of the original
        file_path, file_name = path.split(self.get_read_diagram().get_file_path())
        file_name, file_ext = file_name.split('.')
        new_file_name = f"{file_name}-errorCheck.{file_ext}"
        tree.write(path.join(file_path, new_file_name))

    def get_read_diagram(self):
        return self._read_diagram

    def _get_print_id_width(self):
        return self._print_id_width

    def _get_separator_width(self):
        return self._separator_width


class WrongFileFormatError(Exception):
    def __init__(self, message="Wrong file format provided"):
        self.message = message
        super().__init__(self.message)

class DisconnectedTermError(Exception):
    def __init__(self, term_id, term_content):
        if not term_content.strip():
            self.message = f"Term with id {term_id} is not connected to any other term."
        else:
            self.message = f"Term with content: {term_content}, is not connected to any other term."
        super().__init__(self.message)

class DisconnectedEdgeError(Exception):
    def __init__(self, message):
        super().__init__(message)

class MissingParentEdgeError(DisconnectedEdgeError):
    def __init__(self, edge_id, edge_content):
        if not edge_content.strip():
            self.message = f"Edge with id {edge_id} does not have a source (parent) connected."
        else:
            self.message = f"Edge with content: {edge_content}, does not have a source (parent) connected."
        super().__init__(self.message)

class MissingChildEdgeError(DisconnectedEdgeError):
    def __init__(self, edge_id, edge_content):
        if not edge_content.strip():
            self.message = f"Edge with id {edge_id} does not have a target (child) connected."
        else:
            self.message = f"Edge with content: {edge_content}, does not have a target (child) connected."
        super().__init__(self.message)

class FloatingEdgeError(DisconnectedEdgeError):
    def __init__(self, edge_id, edge_content):
        if not edge_content:
            self.message = f"Edge with id {edge_id} does not have anything connected."
        else:
            self.message = f"Edge with content: {edge_content}, does not have anything connected."
        super().__init__(self.message)

class CircularEdgeError(DisconnectedEdgeError):
    def __init__(self, edge_id, edge_content):
        if not edge_content:
            self.message = f"Edge with id {edge_id} is only connected to itself. Please ignore."
        else:
            self.message = f"Edge with content: {edge_content}, is only connected to itself. Please ignore."
        super().__init__(self.message)

class BlankLabelError(Exception):
    def __init__(self, message):
        super().__init__(message)

class BlankTermLabelError(BlankLabelError):
    def __init__(self, term_id):
        self.message = f"Term with id {term_id} does not have a label."
        super().__init__(self.message)

class BlankEdgeLabelError(BlankLabelError):
    def __init__(self, edge_id, connected_terms):
        if not connected_terms:
            self.message = f"Edge with id {edge_id} does not have a label."
        else:
            self.message = f"Edge connected to {' and '.join(connected_terms)}, does not have a label."
        super().__init__(self.message)



import sys
from PyQt5.QtWidgets import (
    QApplication, QWidget, QLabel, QLineEdit, QPushButton,
    QVBoxLayout, QHBoxLayout, QFileDialog, QTableWidget, QTableWidgetItem, QHeaderView, QSpacerItem, QSizePolicy, QProgressDialog
)
from PyQt5.QtCore import Qt, QThread, pyqtSignal

from cemento.draw_io.read_diagram import ReadDiagram

class FileParserApp(QWidget):
    def __init__(self):
        super().__init__()

        self.initUI()

    def initUI(self):
        # Layouts
        main_layout = QVBoxLayout()
        file_layout = QHBoxLayout()
        output_layout = QHBoxLayout()
        button_layout = QHBoxLayout()

        # Input File Selection
        self.input_label = QLabel("Input File:")
        self.input_path = QLineEdit()
        self.input_path.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.input_button = QPushButton("Browse")
        self.input_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.input_button.clicked.connect(self.select_input_file)
        file_layout.addWidget(self.input_label)
        file_layout.addWidget(self.input_path)
        file_layout.addWidget(self.input_button)

        # Output File Selection
        self.output_label = QLabel("Output File:")
        self.output_path = QLineEdit()
        self.output_path.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.output_button = QPushButton("Browse")
        self.output_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.output_button.clicked.connect(self.select_output_file)
        output_layout.addWidget(self.output_label)
        output_layout.addWidget(self.output_path)
        output_layout.addWidget(self.output_button)

        # Parse Button
        self.parse_button = QPushButton("Read Diagram")
        self.parse_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.parse_button.clicked.connect(self.start_parsing)
        button_layout.addItem(QSpacerItem(20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        button_layout.addWidget(self.parse_button)
        button_layout.addItem(QSpacerItem(20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # Error Table
        self.error_table = QTableWidget()
        self.error_table.setColumnCount(2)
        self.error_table.setHorizontalHeaderLabels(["Element ID", "Error Description"])
        self.error_table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        # Adding layouts to main layout
        main_layout.addLayout(file_layout)
        main_layout.addLayout(output_layout)
        main_layout.addLayout(button_layout)
        main_layout.addWidget(self.error_table)

        self.setLayout(main_layout)
        self.setWindowTitle("Cemento (Read your Ontologies)")
        self.setGeometry(300, 300, 600, 400)

    def select_input_file(self):
        file_name, _ = QFileDialog.getOpenFileName(self, "Select Input File")
        if file_name:
            self.input_path.setText(file_name)

    def select_output_file(self):
        file_name, _ = QFileDialog.getSaveFileName(self, "Select Output File")
        if file_name:
            self.output_path.setText(file_name)

    def start_parsing(self):
        input_file = self.input_path.text()
        output_file = self.output_path.text()

        if not input_file or not output_file:
            self.show_error("Please select both input and output paths.")
            return

        self.progress_dialog = QProgressDialog("Parsing file...", "Cancel", 0, 100, self)
        self.progress_dialog.setWindowModality(Qt.WindowModal)
        self.progress_dialog.setMinimumDuration(0)
        self.progress_dialog.setAutoClose(True)

        self.parser_thread = ParserThread(input_file, output_file)
        self.parser_thread.progress_update.connect(self.progress_dialog.setValue)
        self.parser_thread.finished.connect(self.on_parsing_finished)
        self.progress_dialog.canceled.connect(self.parser_thread.terminate)

        self.parser_thread.start()

    def on_parsing_finished(self):
        self.populate_error_table(self.parser_thread.errors)
        self.progress_dialog.reset()

    def parse_file(self, progress_callback):
        # Example dummy parsing logic that generates errors
        errors = []
        for i in range(1, 101):
            progress_callback.emit(i)
            if i % 20 == 0:
                errors.append((i, f"Error at line {i}"))

        return errors

    def populate_error_table(self, errors):
        self.error_table.setRowCount(0)
        for element_id, description in errors:
            row_position = self.error_table.rowCount()
            self.error_table.insertRow(row_position)

            if not element_id:
                element_id = "N/A"

            self.error_table.setItem(row_position, 0, QTableWidgetItem(str(element_id)))
            self.error_table.setItem(row_position, 1, QTableWidgetItem(description))

    def show_error(self, message):
        self.error_table.setRowCount(0)
        row_position = self.error_table.rowCount()
        self.error_table.insertRow(row_position)
        self.error_table.setItem(row_position, 0, QTableWidgetItem("N/A"))
        self.error_table.setItem(row_position, 1, QTableWidgetItem(message))

class ParserThread(QThread):
    progress_update = pyqtSignal(int)
    
    def __init__(self, input_file_path, output_file_path):
        super().__init__()
        self._input_file_path = input_file_path
        self._output_file_path = output_file_path
        self.errors = []

    def _get_input_file_path(self):
        return self._input_file_path
    
    def _get_output_file_path(self):
        return self._output_file_path
    
    def _add_errors(self, error_record):
        self.errors.append(error_record)
    
    def run(self):
        try:
            read_diagram = ReadDiagram(file_path=self._get_input_file_path())
            self.progress_update.emit(33)

            rels_df = read_diagram.get_relationships()
            rels_df.to_csv(self._get_output_file_path())
            self.progress_update.emit(66)

            read_errors = read_diagram.get_errors()
            for error_id, error_data in read_errors.items():
                for error in error_data['errors']:
                    self._add_errors((error_id, str(error)))
        except Exception as e:
            self._add_errors((None, str(e)))

if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = FileParserApp()
    ex.show()
    sys.exit(app.exec_())


import networkx as nx
from cemento.graph import Graph


class Tree(Graph):

    def __init__(
        self,
        file_path=None,
        rels_df=None,
        graph=None,
        ref=None,
        do_gen_ids=False,
        infer_rank=False,
        invert_tree=False
    ):
        super().__init__(
            file_path=file_path,
            rels_df=rels_df,
            graph=graph,
            ref=ref,
            do_gen_ids=do_gen_ids,
            infer_rank=infer_rank
        )
        self._root = None
        self._invert_tree = invert_tree

    def get_subgraphs(self, input_tree=None):
        if not input_tree:
            input_tree = self.get_graph()

        subgraphs = [
            input_tree.subgraph(c).copy()
            for c in nx.weakly_connected_components(input_tree)
        ]
        subgraphs = [Tree(graph=subgraph, ref=self.get_ref(), invert_tree=self.get_invert_tree()) for subgraph in subgraphs]
        return subgraphs

    def _compute_grid_allocs(self):
        for node in self.get_nodes():
            set_reserved_x = 1 if self._graph.out_degree(node) == 0 else 0
            self.set_attr(node, "reserved_x", set_reserved_x)
            self.set_attr(node, "reserved_y", 1)

        for node in reversed(list(nx.bfs_tree(self.get_graph(), self.get_root()))):
            if len(nx.descendants(self.get_graph(), node)) > 0:
                max_reserved_y = 0
                for child in self._graph.successors(node):
                    new_reserved_x = self.get_attr(node, "reserved_x") + self.get_attr(
                        child, "reserved_x"
                    )
                    max_reserved_y = max(
                        max_reserved_y, self.get_attr(child, "reserved_y")
                    )
                    self.set_attr(node, "reserved_x", new_reserved_x)

                new_reserved_y = max_reserved_y + self.get_attr(node, "reserved_y")
                self.set_attr(node, "reserved_y", new_reserved_y)

    def _compute_draw_pos(self):
        for level, nodes_in_level in enumerate(
            nx.bfs_layers(self.get_graph(), self.get_root())
        ):
            for node in nodes_in_level:
                # assign y-level to current level
                self.set_attr(node, "draw_y", level)

        # start cursor at a zero-origin
        self.set_attr(self.get_root(), "cursor_x", 0)
        # iterate over nodes in dfs to assign cursor positions
        for node in nx.dfs_preorder_nodes(self.get_graph(), self.get_root()):
            offset_x = 0  # re-initialize offset relative to sibling to zero
            cursor_x = self.get_attr(node, "cursor_x")  # get the current x-cursor

            # iterate over children to assign cursors with inherited parent cursor
            for child in self._graph.successors(node):
                child_cursor_x = (
                    cursor_x + offset_x
                )  # inherit parent cursor and add offset relative to already-drawn siblings
                self.set_attr(child, "cursor_x", child_cursor_x)
                offset_x += self.get_attr(child, "reserved_x")

            self.set_attr(
                node, "draw_x", (2 * cursor_x + self.get_attr(node, "reserved_x")) / 2
            )  # assign draw position at midpoint between x-span and cursor

        if self.get_invert_tree():
            for node in self.get_nodes():
                draw_x = self.get_attr(node, 'draw_x')
                draw_y = self.get_attr(node, 'draw_y')

                self.set_attr(node, 'draw_x', draw_y)
                self.set_attr(node, 'draw_y', draw_x)

    def draw_tree(
        self, write_diagram, translate_x=0, translate_y=0, draw_predicates=True
    ):

        term_uids = dict()
        subgraph_sizes = dict()
        ranked_graph = self._create_graph(self.get_rank_edges())
        ranked_subgraphs = self.get_subgraphs(input_tree=ranked_graph)

        for subgraph in ranked_subgraphs:
            # compute draw positions
            write_diagram.update_graph_count(1)
            current_tree_ct = write_diagram.get_graph_count()

            subgraph._compute_grid_allocs()
            subgraph._compute_draw_pos()

            # draw each node
            for term_id in subgraph.get_nodes():
                term_content = subgraph._ref.get_term(term_id)
                pos_x, pos_y = (
                    subgraph.get_attr(term_id, "draw_x") + translate_x,
                    subgraph.get_attr(term_id, "draw_y") + translate_y,
                )
                self.set_attr(term_id, "pos_x", pos_x)
                self.set_attr(term_id, "pos_y", pos_y)
                self.set_attr(term_id, "tree_number", current_tree_ct)
                term_uids[term_id] = write_diagram.add_shape(term_content, pos_x, pos_y)

            for edge in subgraph.get_edges():
                parent_id, child_id = edge
                rel_content = subgraph._ref.get_rel_from_edge(edge)
                write_diagram.add_connector(
                    term_uids[parent_id],
                    term_uids[child_id],
                    rel_content,
                    is_rank=True,
                    inverted=self.get_invert_tree()
                )

            subgraph_size = subgraph.get_size()
            subgraph_sizes[write_diagram.get_graph_count()] = subgraph_size
            if self.get_invert_tree():
                translate_x += subgraph_size[1]
            else:
                translate_x += subgraph_size[0]

        if draw_predicates:
            self._draw_predicates(term_uids, write_diagram, subgraph_sizes)

    def _draw_predicates(self, term_uids, write_diagram, subgraph_sizes):
        y_offset = dict()

        for edge in self.get_predicate_edges():
            parent_id, child_id, _ = edge
            drawn_ids = set(term_uids.keys())

            if parent_id not in drawn_ids and child_id not in drawn_ids:
                continue
            elif parent_id not in drawn_ids or child_id not in drawn_ids:
                created_id = parent_id if parent_id in drawn_ids else child_id
                uncreated_id = child_id if child_id not in drawn_ids else parent_id
                # add created_id to dictionary if not already there
                if created_id not in y_offset.keys():
                    y_offset[created_id] = 0
                # compute new term parameters
                uncreated_content = self._ref.get_term(uncreated_id)
                pos_x = self.get_attr(created_id, "pos_x")
                tree_end_y = subgraph_sizes[self.get_attr(created_id, "tree_number")][1]
                pos_y = tree_end_y + y_offset[created_id]
                # create new shape for term and add to ids
                self.add_node(uncreated_id)
                self.set_attr(uncreated_id, "pos_x", pos_x)
                self.set_attr(uncreated_id, "pos_y", pos_y)
                term_uids[uncreated_id] = write_diagram.add_shape(
                    uncreated_content, pos_x, pos_y
                )
                y_offset[created_id] += 0.5

            # create new connector
            edge = (parent_id, child_id) = edge[:2] # only key with the id tuple
            rel_content = self._ref.get_rel_from_edge(edge)

            write_diagram.add_connector(
                term_uids[parent_id],
                term_uids[child_id],
                rel_content,
                is_rank=False,
                inverted=self.get_invert_tree()
            )

    def get_invert_tree(self):
        return self._invert_tree

    def get_root(self):
        if not self._root:
            try:
                self._root = next(nx.topological_sort(self._graph))
            except Exception:
                self._root = None
                return self._root
        return self._root

    def get_size(self):
        root = self.get_root()

        if not self.get_attr(root, "reserved_x") or not self.get_attr(root, "reserved_y"):
            self._compute_grid_allocs()

        return (self.get_attr(root, "reserved_x"), self.get_attr(root, "reserved_y"))


from collections import defaultdict
import math
import networkx as nx
from cemento.graph import Graph

# TODO: move to graph ref
DEFAULT_LEVEL = 4

class CircleGraph(Graph):

    def __init__(
        self,
        file_path=None,
        rels_df=None,
        graph=None,
        ref=None,
        do_gen_ids=False,
        node_distance=20,
        scale=1000,
        center=(0, 0),
        iterations=1000,
        min_edge_weight = 10,
        max_edge_weight = 50
    ):
        super().__init__(
            file_path=file_path,
            rels_df=rels_df,
            graph=graph,
            ref=ref,
            do_gen_ids=do_gen_ids,
            infer_rank=False
        )
        self._node_distance = node_distance
        self._scale = scale
        self._center = center
        self._iterations = iterations
        self._min_edge_weight = min_edge_weight
        self._max_edge_weight = max_edge_weight

    def compute_draw_pos(self):
        # Compute the positions of nodjes using the spring_layout function
        node_level = defaultdict(lambda: DEFAULT_LEVEL)
        max_level = 0
        edges = [(u, v) for u, v, rel_type in self.get_edges() if rel_type == 'rank']
        temp_graph = nx.DiGraph(edges)
        temp_subgraphs = [temp_graph.subgraph(c).copy() for c in nx.weakly_connected_components(temp_graph)]
        for subgraph in temp_subgraphs:
            root = next(nx.topological_sort(subgraph))
            levels = nx.single_source_shortest_path_length(subgraph, root)
            for node, level in levels.items():
                node_level[node] = level
                max_level = max(max_level, level)

        input_graph = self._graph.to_undirected()

        for edge_pair in input_graph.edges():
            u, v = edge_pair
            edge_weight = (node_level[u] - node_level[v])

            min_weight = self.get_min_edge_weight()
            max_weight = self.get_max_edge_weight()

            edge_weight = (min_weight/math.sqrt(max_weight/min_weight)) * (math.sqrt(50/10)**edge_weight)
            nx.set_edge_attributes(input_graph, {(u, v): {'weight': edge_weight}})

        positions = nx.spring_layout(
            input_graph,
            k=self.get_node_distance(),
            scale=self.get_scale(),
            center=self.get_center(),
            iterations=self.get_iterations()
        )

        # Assign the x and y coordinates as attributes to each node
        for node, pos in positions.items():
            self.set_attr(node, "draw_x", pos[0])
            self.set_attr(node, "draw_y", pos[1])

        return positions

    def draw_graph(self, write_diagram, positions=None):
        term_uids = dict()

        if not positions:
            self.compute_draw_pos()

        for term_id in self.get_nodes():
            term_content = self._ref.get_term(term_id)
            pos_x, pos_y = self.get_attr(term_id, "draw_x"), self.get_attr(term_id, "draw_y")
            term_uids[term_id] = write_diagram.add_circle(term_content, pos_x, pos_y, apply_scale=False)

        for edge in self.get_edges():
            parent_id, child_id, _ = edge
            rel_content = self._ref.get_rel_from_edge((parent_id, child_id))

            parent_x, parent_y = self.get_attr(parent_id, 'draw_x'), self.get_attr(parent_id, 'draw_y')
            child_x, child_y = self.get_attr(child_id, 'draw_x'), self.get_attr(child_id, 'draw_y')
            angle = math.atan2((child_y - parent_y), (child_x - parent_x))
            start_x, start_y = CircleGraph.get_square_coordinate(angle)
            end_x, end_y = CircleGraph.get_square_coordinate(angle)

            write_diagram.add_straight_connector(
                term_uids[parent_id],
                term_uids[child_id],
                rel_content,
                start_pos_x=start_x,
                start_pos_y=start_y,
                end_pos_x=end_x,
                end_pos_y=end_y
            )

    def get_node_distance(self):
        return self._node_distance

    def set_node_distance(self, spring_constant):
        self._node_distance = spring_constant
        self.compute_draw_pos()

    def get_scale(self):
        return self._scale

    def set_scale(self, scale):
        self._scale = scale
        self.compute_draw_pos()

    def get_center(self):
        return self._center

    def set_center(self, center_x, center_y):
        self._center = (center_x, center_y)
        self.compute_draw_pos()

    def get_iterations(self):
        return self._iterations

    def set_iterations(self, iterations):
        self._iterations = iterations
        self.compute_draw_pos()

    def get_min_edge_weight(self):
        return self._min_edge_weight

    def get_max_edge_weight(self):
        return self._max_edge_weight

    @staticmethod
    def get_square_coordinate(angle, translate_x=0, translate_y=0):
        return ((math.cos(angle) + translate_x) * 0.5 + 0.5, (math.sin(angle) + translate_y) * 0.5 + 0.5)


from os import path

class Diagram:

    def __init__(self, file_path):

        self._file_path = file_path
        self._name = None
        self._modify_date = None
        self._diagram_id = None
        self._grid_dx = None
        self._grid_dy = None
        self._grid_size = None
        self._page_width = None
        self._page_height = None

        self._name = self._set_name()

    def set_file_path(self, file_path):
        self._file_path = file_path
        self._set_name()

    def get_file_path(self):
        return self._file_path

    def _set_name(self):
        self._name = path.basename(self._file_path)

    def get_name(self):
        return self._name

    def get_modify_date(self):
        return self._modify_date

    def set_modify_date(self, modify_date):
        self._modify_date = modify_date

    def get_diagram_id(self):
        return self._diagram_id

    def get_grid_dx(self):
        return self._grid_dx

    def set_grid_dx(self, grid_dx):
        self._grid_dx = grid_dx

    def get_grid_dy(self):
        return self._grid_dy

    def set_grid_dy(self, grid_dy):
        self._grid_dy = grid_dy

    def get_grid_size(self):
        return self._grid_size

    def set_grid_size(self, grid_size):
        self._grid_size = grid_size

    def get_page_width(self):
        return self._page_width

    def set_page_width(self, page_width):
        self._page_width = page_width

    def get_page_height(self):
        return self._page_height

    def set_page_height(self, page_height):
        self._page_height = page_height


import pandas as pd


class WriteArray:

    def __init__(self, read_diagram):

        if not read_diagram._parse_terms:
            raise AttributeError(
                "Cannot write output if terms are not parsed. Please set parse_terms arg to True in ReadDiagram."
            )

        self._diagram = read_diagram
        self._rels = read_diagram.get_relationships()
        self._terms_info = read_diagram.get_terms_info()

        self._var_array = None
        self._rel_array = None

        self._set_var_array()
        self._set_rel_array()

    def _prepare_terms(self, titles=None, get_terms=True):
        if not titles:
            titles = self.get_diagram().get_diagram_ref().get_term_parser_titles()

        info_df = self.get_terms_info().copy()
        info_df = (
            info_df[info_df["is_term"]] if get_terms else info_df[~info_df["is_term"]]
        )

        # ensure all the title columns are made if not in the dataframe already
        for title in titles.values():
            if title not in info_df.columns:
                info_df[title] = pd.Series([])

        return info_df

    def _set_var_array(self):
        titles = self.get_diagram().get_diagram_ref().get_term_parser_titles()
        term_info = self._prepare_terms(titles=titles, get_terms=True)

        export_colname_map = {
            "variable": "VariableName",
            "prefix": "WhichOntologyItBelongsTo",
            "parent": "ParentVariable",
            "parent_prefix": "WhichOntologyParentBelongsTo",
            titles["definition_title"]: "Definition",
            titles["alt_names_title"]: "AlternativeNames",
            titles["unit_title"]: "Unit",
            titles["axioms_title"]: "LogicalAxioms",
            titles["unit_onto_title"]: "WhichOntoUnitBelongsTo",
        }
        remove_cols = [
            col for col in term_info.columns if col not in export_colname_map.keys()
        ]
        term_info.drop(remove_cols, axis=1, inplace=True)
        term_info.rename(columns=export_colname_map, inplace=True)

        self._var_array = term_info

    def _set_rel_array(self):
        titles = self.get_diagram().get_diagram_ref().get_term_parser_titles()
        rel_info = self._prepare_terms(titles=titles, get_terms=False)

        export_colname_map = {
            "variable": "RelationshipName",
            "prefix": "WhichOntologyItBelongsTo",
            "subject": "Subject",
            "object": "Object",
            "is_data_prop": "IsDataProperty",
            titles["definition_title"]: "Definition",
            titles["alt_names_title"]: "AlternativeNames",
            titles["axioms_title"]: "LogicalAxioms",
            titles["data_type_title"]: "DataType",
        }
        remove_cols = [
            col for col in rel_info.columns if col not in export_colname_map.keys()
        ]
        rel_info.drop(remove_cols, axis=1, inplace=True)
        rel_info.rename(columns=export_colname_map, inplace=True)

        self._rel_array = rel_info

    def get_diagram(self):
        return self._diagram

    def get_rels(self):
        return self._rels

    def get_terms_info(self):
        return self._terms_info

    def get_var_array(self):
        return self._var_array

    def get_rel_array(self):
        return self._rel_array



import os
import importlib.resources as pkg_resources
from pathlib import Path
from configparser import ConfigParser
from string import Template

class DiagramRef:

    def __init__(self):
        self._templates = None
        self._term_colors = None
        self._diagram_defaults = None
        self._rank_list = None
        self._connector_list = None
        self._shape_params = None
        self._term_parser_titles = None
        self._term_parser_sep = None
        self._config = None

        self._read_config()
        self._read_templates()
        self._read_color_dict()
        self._read_defaults()
        self._read_rank_list()
        self._read_instance_connector_list()
        self._read_shape_params()
        self._read_term_parser_titles()
        self._read_term_parser_sep()

    def _read_config(self):
        config = ConfigParser()

        current_file_folder = Path(__file__)
        # retrieve configuration file from the grandparent directory
        config_path = current_file_folder.parent.parent.parent /  "config.ini"

        if config_path.exists():
            config.read(config_path)
        else:
            with pkg_resources.open_text('cemento', 'config.ini') as config_file:
                config.read_file(config_file)

        self._config = config

    def _read_templates(self):
        current_file_folder = Path(__file__)
        # retrieve the template folder from the grandparent directory
        template_path = current_file_folder.parent.parent.parent / "templates"

        if not template_path.exists():
            template_path = pkg_resources.files('cemento').joinpath('templates')

        template_files = [file for file in os.scandir(template_path) if file.name.endswith('.xml')]
        templates = {file.name.replace('.xml', ''): Template(open(file.path).read()) for file in template_files}
        self._templates = templates

    def _read_color_dict(self):
        config = self._get_config()
        color_dict = dict(config.items('term_colors'))
        self._term_colors = color_dict

    def _read_defaults(self):
        config = self._get_config()
        diagram_defaults = dict(config.items('diagram_defaults'))
        diagram_defaults = {key: int(value) for key, value in diagram_defaults.items()}
        self._diagram_defaults = diagram_defaults

    def _read_rank_list(self):
        config = self._get_config()
        sep = config['rank_terms']['sep']
        rank_list = config['rank_terms']['term_list'].split(sep)
        self._rank_list = [rank_term.lower() for rank_term in rank_list]

    def _read_instance_connector_list(self):
        config = self._get_config()
        sep = config['rank_terms']['sep']
        connector_list = config['rank_terms']['instance_connectors'].split(sep)
        self._connector_list = [rank_term.lower() for rank_term in connector_list]

    def _read_term_parser_titles(self):
        config = self._get_config()
        term_titles = dict(config.items('term_parser'))
        term_titles = {key.strip(): value.strip() for key, value in term_titles.items() if 'title' in key}
        self._term_parser_titles = term_titles

    def _read_term_parser_sep(self):
        config = self._get_config()
        term_sep = config['term_parser']['sep']
        self._term_parser_sep = term_sep

    def _read_shape_params(self):
        config = self._get_config()
        shape_param_dict = dict(config.items('shape_params'))
        shape_param_dict = {key: int(value) if value.isdigit() else value for key, value in shape_param_dict.items()}

        self._shape_params = shape_param_dict

    def _get_config(self):
        return self._config

    def get_term_color(self, term):
        if not self._term_colors:
            self._read_color_dict()

        if term not in self._term_colors:
            term = 'default'

        return self._term_colors[term]

    def get_template(self, key):
        if not self._templates:
            self._read_templates()
        return self._templates[key]

    def get_diagram_defaults(self, key):
        if not self._diagram_defaults:
            self._read_defaults()
        return self._diagram_defaults[key]

    def get_term_parser_title(self, key):
        if not self._term_parser_titles:
            self._read_term_parser_titles()
        return self._term_parser_titles[key]

    def get_term_parser_sep(self):
        return self._term_parser_sep

    def get_term_parser_titles(self):
        return self._term_parser_titles

    def is_rank(self, term):
        return term.lower() in self._rank_list

    def is_instance_connector(self, term):
        return term.lower() in self._connector_list

    def get_shape_param(self, key):
        if not self._shape_params:
            self._read_shape_params()
        return self._shape_params[key]